from typing import Any

from .query_parameter import QueryParameter, QueryParams


class ParameterBuilder:
    def __init__(self, *parameter_types: type[QueryParameter]) -> None:
        self._parameter_types = parameter_types

    def check_query_parameters(self, args: dict[str, Any]) -> None:
        """Check query parameters.

        Parameters
        ----------
        args : dict[str, Any]
            Arguments passed as query parameters.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        """
        errors: list[str] = []
        for parameter in self._parameter_types:
            parameter.check_query_parameters(errors, args)

        if errors:
            raise ValueError(", ".join(errors))

    def build_query_parameters(
        self, args: dict[str, Any], extra_params: QueryParams | None
    ) -> QueryParams:
        """Build query parameters."""
        parameters: QueryParams = {}

        if extra_params is not None:
            parameters = extra_params.copy()

        for parameter in self._parameter_types:
            parameter.create_query_parameters(parameters, args)
        return parameters



from enum import Enum


class TargetType(Enum):
    """Enumeration of target types."""

    POWER_SERVICE_LOCATION = "POWER_SERVICE_LOCATION"
    """
    A Power Service Location is a utility named specific location in geography or
    the distribution system, usually the point of service to a customer site.
    """
    SERVICE_AREA = "SERVICE_AREA"
    """
    A Service Area is a utility named geographic region.
    Target values array contains a string representing a service area name.
    """
    GROUP = "GROUP"
    """
    Target values array contains a string representing a group.
    """
    RESOURCE_NAME = "RESOURCE_NAME"
    """
    Target values array contains a string representing a resource name.
    """
    VEN_NAME = "VEN_NAME"
    """
    Target values array contains a string representing a VEN name.
    """
    EVENT_NAME = "EVENT_NAME"
    """
    Target values array contains a string representing an event name.
    """
    PROGRAM_NAME = "PROGRAM_NAME"
    """
    Target values array contains a string representing a program name.
    """



from .docstringbasemodel import DocstringBaseModel
from .intervalperiod import IntervalPeriod
from .valuesmap import ValuesMap


class Interval(DocstringBaseModel):
    """Interval object.

    Has a payload for an IntervalPeriod. The IntervalPeriod is either local to the class or
    available from the parent Event object.
    """

    id: int
    """The ID of the interval."""

    interval_period: IntervalPeriod | None = None
    """The interval period of the interval."""

    payloads: list[ValuesMap]
    """The payloads of the interval."""

    def has_interval_period(self) -> bool:
        """Check if the interval has an interval period."""
        return self.interval_period is not None



import datetime
from typing import Any

from pydantic import field_validator
from pydantic_core.core_schema import ValidationInfo

from .docstringbasemodel import DocstringBaseModel
from .isoduration import parse_iso8601_duration


class IntervalPeriod(DocstringBaseModel):
    """Interval period object.

    Represents a period of time with a start time, duration, and randomize start time.

    P9999Y durations are supported by sending the `allow_P9999Y_duration` flag in the context.
    For example:

    >>> IntervalPeriod.model_validate(data, context={"allow_P9999Y_duration": True})

    The `0000-00-00` start date is supported by sending the `0000-00-00` flag in the context.
    For example:

    >>> IntervalPeriod.model_validate(data, context={"0000-00-00": datetime.datetime(2024, 9, 24)})
    """

    # TODO @jepebe: if there is no time zone data, the Program object may have a timeZoneOffset
    #      that can be used to adjust the time to the correct timezone
    #      https://github.com/prelectai/toadr3/issues/23
    start: datetime.datetime
    """The start of the interval period."""

    duration: datetime.timedelta = datetime.timedelta(seconds=0)
    """The duration of the interval period."""

    randomize_start: datetime.timedelta = datetime.timedelta(seconds=0)
    """The randomize start of the interval period."""

    @field_validator("duration", "randomize_start", mode="before")
    @classmethod
    def validate_timedeltas(
        cls, iso_duration: str | None, info: ValidationInfo
    ) -> datetime.timedelta:
        """Convert the iso duration to a timedelta object."""
        if iso_duration is None:
            return datetime.timedelta(seconds=0)

        if info.context is not None:
            allow = info.context.get("allow_P9999Y_duration", False)
            if allow and iso_duration == "P9999Y":
                return datetime.timedelta(days=365 * 9999)  # close enough

        return parse_iso8601_duration(iso_duration)

    @field_validator("start", mode="before")
    @classmethod
    def validate_start(cls, start: Any, info: ValidationInfo) -> Any:  # noqa: ANN401
        """Handle 0000-00-00 if configured."""
        if start == "0000-00-00" and info.context is not None and "0000-00-00" in info.context:
            return info.context["0000-00-00"]

        return start

    def __str__(self) -> str:
        """Return a description of the interval period."""
        return f"{self.start} - {self.start + self.duration} (± {self.randomize_start})"



import aiohttp

from ._internal import (
    ClientName,
    EventID,
    ParameterBuilder,
    ProgramID,
    SkipAndLimit,
    default_error_handler,
    get_query,
)
from .access_token import AccessToken
from .models import Report

_GET_PARAMS_BUILDER = ParameterBuilder(ProgramID, EventID, ClientName, SkipAndLimit)


async def post_report(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    report: Report,
    custom_headers: dict[str, str] | None = None,
) -> Report:
    """Post a report to the VTN.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url : str
        The URL of the VTN.
    access_token : AccessToken | None
        The access token to use for the request, use None if no token is required.
    report : Report
        The report object to post.
    custom_headers : dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    Report
        The report object returned by the VTN.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrError
        If the request to the VTN fails. Specifically, if the
        response status is 400, 403, 409, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.
    """
    if report is None:
        raise ValueError("report is required")

    headers: dict[str, str] = {}
    if custom_headers is not None:
        headers |= custom_headers

    if access_token is not None:
        headers["Authorization"] = f"Bearer {access_token.token}"

    vtn_url = vtn_url.rstrip("/")

    data = report.model_dump_json(exclude_none=True, exclude_unset=True)
    headers["Content-Type"] = "application/json"

    async with session.post(f"{vtn_url}/reports", headers=headers, data=data) as response:
        if not response.ok:
            match response.status:
                case 400 | 403 | 409 | 500:
                    await default_error_handler(response)
                case _:
                    await default_error_handler(response, "Unexpected error status!")

        data = await response.json()
        return Report.model_validate(data)


async def get_reports(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    program_id: str | None = None,
    event_id: str | None = None,
    client_name: str | None = None,
    skip: int | None = None,
    limit: int | None = None,
    extra_params: dict[str, str | int | list[str]] | None = None,
    custom_headers: dict[str, str] | None = None,
) -> list[Report]:
    """Get a list of reports from the VTN.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url : str
        The URL of the VTN.
    access_token : AccessToken | None
        The access token to use for the request, use None if no token is required.
    program_id : str | None
        The program ID to filter the reports by.
    event_id : str | None
        The event ID to filter the reports by.
    client_name : str | None
        The client name to filter the reports by.
    skip : int | None
        The number of reports to skip (for pagination).
    limit : int | None
        The maximum number of reports to return.
    extra_params : dict[str, str | int | list[str]] | None
        Extra query parameters to include in the request.
    custom_headers : dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    list[Report]
        A list of report objects.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrException
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.
    """
    args = {
        "client_name": client_name,
        "event_id": event_id,
        "program_id": program_id,
        "skip": skip,
        "limit": limit,
    }
    _GET_PARAMS_BUILDER.check_query_parameters(args)
    params = _GET_PARAMS_BUILDER.build_query_parameters(args, extra_params)
    data = await get_query(session, f"{vtn_url}/reports", access_token, params, custom_headers)

    if not isinstance(data, list):
        raise ValueError(f"Expected result to be a list. Got {type(data)} instead.")

    result = []
    for report in data:
        result.append(Report.model_validate(report))
    return result



from enum import StrEnum

from toadr3.models import DocstringBaseModel


class ObjectType(StrEnum):
    """Types of objects addressable through API."""

    EVENT = "EVENT"
    PROGRAM = "PROGRAM"
    REPORT = "REPORT"
    RESOURCE = "RESOURCE"
    SUBSCRIPTION = "SUBSCRIPTION"
    VEN = "VEN"


class OperationType(StrEnum):
    """Object operation to subscribe to."""

    DELETE = "DELETE"
    GET = "GET"
    POST = "POST"
    PUT = "PUT"


class ObjectOperation(DocstringBaseModel):
    """Object type, operations, and callbackUrl."""

    objects: list[ObjectType]
    """List of objects to subscribe to."""

    operations: list[OperationType]
    """List of operations to subscribe to."""

    callback_url: str
    """User provided webhook URL."""

    bearer_token: str | None = None
    """User provided token.

    To avoid custom integrations, callback endpoints should accept the provided bearer
    token to authenticate VTN requests.
    """



from typing import Any

from .query_parameter import QueryParameter, QueryParams


class SkipAndLimit(QueryParameter):
    """Skip and limit parameters.

    skip : int | None
        The number of items to skip (for pagination).
    limit : int | None
        The maximum number of items to return.
    """

    @staticmethod
    def create_query_parameters(params: QueryParams, args: dict[str, Any]) -> None:
        """Add skip and limit parameters."""
        if "skip" in args and args["skip"] is not None:
            params["skip"] = args["skip"]

        if "limit" in args and args["limit"] is not None:
            params["limit"] = args["limit"]

    @staticmethod
    def check_query_parameters(errors: list[str], args: dict[str, Any]) -> None:
        """Check if skip and limit parameters are valid."""
        skip = args.get("skip")
        limit = args.get("limit")

        if skip is not None:
            if not isinstance(skip, int):
                errors.append("skip must be an integer")
            elif skip < 0:
                errors.append("skip must be a positive integer")

        if limit is not None:
            if not isinstance(limit, int):
                errors.append("limit must be an integer")
            elif limit < 0:
                errors.append("limit must be a positive integer")



import asyncio
from types import TracebackType
from typing import Literal, Self

from aiohttp import ClientSession

import toadr3
from toadr3.models import Event, ObjectType, Program, Report, Subscription, TargetType


class ToadrClient:
    """Client to interact with OpenADR3 VTN servers.

    If the client created its own session during initialization, the client needs to
    be closed before shutdown.
    """

    def __init__(
        self,
        vtn_url: str,
        oauth_config: toadr3.OAuthConfig | None,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize the client.

        Parameters
        ----------
        vtn_url : str
            The URL of the VTN server to connect to.
        oauth_config : toadr3.OAuthConfig | None
            The OAuth configuration to use or None if credentials are not required.
        session : ClientSession | None
            The session to use or None if the client should make its own session.
        """
        self._vtn_url = vtn_url.rstrip("/")
        self._oauth_config = oauth_config
        self._session = session if session is not None else ClientSession()
        self._token_lock = asyncio.Lock()
        self._token: toadr3.AccessToken | None = None
        self._closed = False

    @property
    def client_session(self) -> ClientSession:
        """Client session for the Switch API."""
        return self._session

    @property
    def vtn_url(self) -> str:
        """VTN server URL."""
        return self._vtn_url

    @property
    def closed(self) -> bool:
        """Whether or not the client is closed."""
        return self._closed

    async def _fetch_token(self) -> toadr3.AccessToken | None:
        """Fetch an access token from the OAuth2 server."""
        if self._oauth_config is None:
            return None
        return await toadr3.acquire_access_token_from_config(self._session, self._oauth_config)

    @property
    async def token(self) -> toadr3.AccessToken | None:
        """Access token for the Switch API."""
        async with self._token_lock:
            if self._token is None or self._token.is_expired():
                self._token = await self._fetch_token()

            return self._token

    async def get_events(
        self,
        program_id: str | None = None,
        target_type: TargetType | str | None = None,
        target_values: list[str] | None = None,
        skip: int | None = None,
        limit: int | None = None,
        extra_params: dict[str, str | int | list[str]] | None = None,
        custom_headers: dict[str, str] | None = None,
    ) -> list[Event]:
        """Get a list of events from the VTN.

        Lists all the events available from the VTN.
        The results can be filtered by providing a program ID, and/or target type and target values.

        Parameters
        ----------
        program_id : str | None
            The program ID to filter the events by.
        target_type : TargetType | str | None
            The target type to filter the events by.
        target_values : list[str] | None
            The target values to filter the events by (names of the target type).
        skip : int | None
            The number of events to skip (for pagination).
        limit : int | None
            The maximum number of events to return.
        extra_params : dict[str, str | int | list[str]] | None
            Extra query parameters to include in the request.
        custom_headers : dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        list[Event]
            A list of events.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrError
            If the request to the VTN fails. Specifically, response status 400, 403, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.
        """
        return await toadr3.get_events(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            program_id=program_id,
            target_type=target_type,
            target_values=target_values,
            skip=skip,
            limit=limit,
            extra_params=extra_params,
            custom_headers=custom_headers,
        )

    async def get_programs(
        self,
        target_type: TargetType | str | None = None,
        target_values: list[str] | None = None,
        skip: int | None = None,
        limit: int | None = None,
        extra_params: dict[str, str | int | list[str]] | None = None,
        custom_headers: dict[str, str] | None = None,
    ) -> list[Program]:
        """Get a list of programs from the VTN.

        List all programs known to the server.
        Use skip and pagination query params to limit response size.

        Parameters
        ----------
        target_type: TargetType | str | None
            The target type to filter the programs by.
        target_values: list[str] | None
            The target values to filter the programs by (names of the target type).
        skip: int | None
            The number of programs to skip (for pagination).
        limit: int | None
            The maximum number of programs to return.
        extra_params: dict[str, str | int | list[str]] | None
            Extra query parameters to include in the request.
        custom_headers: dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        list[Program]
            A list of programs.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrError
            If the request to the VTN fails. Specifically, response status 400, 403, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.
        """
        return await toadr3.get_programs(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            target_type=target_type,
            target_values=target_values,
            skip=skip,
            limit=limit,
            extra_params=extra_params,
            custom_headers=custom_headers,
        )

    async def get_subscriptions(
        self,
        program_id: str | None = None,
        client_name: str | None = None,
        target_type: TargetType | str | None = None,
        target_values: list[str] | None = None,
        objects: list[str] | list[ObjectType] | None = None,
        skip: int | None = None,
        limit: int | None = None,
        extra_params: dict[str, str | int | list[str]] | None = None,
        custom_headers: dict[str, str] | None = None,
    ) -> list[Subscription]:
        """List all subscriptions.

        May filter results by programID and clientID as query params.
        May filter results by objects as query param. See objectTypes schema.
        Use skip and pagination query params to limit response size.

        Parameters
        ----------
        program_id : str | None
            The program ID to filter the subscriptions by.
        client_name : str | None
            The client name to filter the subscriptions by.
        target_type: TargetType | str | None
            The target type to filter the subscriptions by.
        target_values: list[str] | None
            The target values to filter the subscriptions by (names of the target type).
        objects: list[str] | list[ObjectType] | None
            The object types to filter the subscriptions by.
        skip: int | None
            The number of subscriptions to skip (for pagination).
        limit: int | None
            The maximum number of subscriptions to return.
        extra_params: dict[str, str | int | list[str]] | None
            Extra query parameters to include in the request.
        custom_headers: dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        list[Subscription]
            A list of subscriptions.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrException
            If the request to the VTN fails. Specifically, response status 400, 403, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.
        """
        return await toadr3.get_subscriptions(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            program_id=program_id,
            client_name=client_name,
            target_type=target_type,
            target_values=target_values,
            objects=objects,
            skip=skip,
            limit=limit,
            extra_params=extra_params,
            custom_headers=custom_headers,
        )

    async def post_subscription(
        self,
        subscription: Subscription,
        custom_headers: dict[str, str] | None = None,
    ) -> Subscription:
        """Create a new subscription.

        Parameters
        ----------
        subscription: Subscription
            The subscription to create.
        custom_headers: dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        Subscription
            The subscription object created by the VTN.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrException
            If the request to the VTN fails. Specifically, response status 400, 403, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.

        """
        return await toadr3.post_subscription(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            subscription=subscription,
            custom_headers=custom_headers,
        )

    async def get_reports(
        self,
        program_id: str | None = None,
        event_id: str | None = None,
        client_name: str | None = None,
        skip: int | None = None,
        limit: int | None = None,
        extra_params: dict[str, str | int | list[str]] | None = None,
        custom_headers: dict[str, str] | None = None,
    ) -> list[Report]:
        """Get a list of reports from the VTN.

        Parameters
        ----------
        program_id : str | None
            The program ID to filter the reports by.
        event_id : str | None
            The event ID to filter the reports by.
        client_name : str | None
            The client name to filter the reports by.
        skip : int | None
            The number of reports to skip (for pagination).
        limit : int | None
            The maximum number of reports to return.
        extra_params : dict[str, str | int | list[str]] | None
            Extra query parameters to include in the request.
        custom_headers : dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        list[Report]
            A list of report objects.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrException
            If the request to the VTN fails. Specifically, response status 400, 403, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.
        """
        return await toadr3.get_reports(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            program_id=program_id,
            event_id=event_id,
            client_name=client_name,
            skip=skip,
            limit=limit,
            extra_params=extra_params,
            custom_headers=custom_headers,
        )

    async def post_report(
        self,
        report: Report,
        custom_headers: dict[str, str] | None = None,
    ) -> Report:
        """Post a report to the VTN.

        Parameters
        ----------
        report : Report
            The report object to post.
        custom_headers : dict[str, str] | None
            Extra headers to include in the request.

        Returns
        -------
        Report
            The report object returned by the VTN.

        Raises
        ------
        ValueError
            If the query parameters are invalid.
        toadr3.ToadrError
            If the request to the VTN fails. Specifically, response status 400, 403, 409, or 500,
        aiohttp.ClientError
            If there is an unexpected error with the HTTP request to the VTN.
        """
        return await toadr3.post_report(
            session=self._session,
            vtn_url=self._vtn_url,
            access_token=await self.token,
            report=report,
            custom_headers=custom_headers,
        )

    async def close(self) -> None:
        """Close the client session."""
        await self._session.close()
        self._closed = True

    async def __aenter__(self) -> Self:
        """Enter async context."""
        if self.closed:
            raise RuntimeError("Client is closed")
        return self

    async def __aexit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Exit async context."""
        await self.close()
        return False



from typing import Any

from toadr3.models import ObjectType

from .query_parameter import QueryParameter, QueryParams

_OBJECT_TYPES = [str(item) for item in ObjectType]


class Objects(QueryParameter):
    """Objects query parameter.

    objects: list[str] | list[ObjectType] | None
        The object types to filter the items by.
    """

    @staticmethod
    def create_query_parameters(params: QueryParams, args: dict[str, Any]) -> None:
        """Add objects parameter."""
        objects = args.get("objects")
        if objects is not None:
            object_names: list[str] = []
            for obj_name in objects:
                if isinstance(obj_name, ObjectType):
                    object_names.append(obj_name.value)
                else:
                    object_names.append(obj_name)
            params["objects"] = object_names

    @staticmethod
    def check_query_parameters(errors: list[str], args: dict[str, Any]) -> None:
        """Check if objects parameter is valid."""
        objects = args.get("objects")
        if objects is not None:
            if not isinstance(objects, list):
                errors.append("objects must be a list of strings or ObjectType")
            else:
                for item in objects:
                    if not isinstance(item, (ObjectType, str)):
                        errors.append(f"object type '{item}' must be of type string or ObjectType")
                    elif str(item) not in _OBJECT_TYPES:
                        errors.append(f"object type '{item}' must be one of {_OBJECT_TYPES}")



from .client_name import ClientName
from .object_id import EventID, ProgramID, SubscriptionID
from .objects import Objects
from .parameter_builder import ParameterBuilder
from .query_handler import default_error_handler, get_query
from .query_parameter import QueryParameter, QueryParams
from .skip_and_limit import SkipAndLimit
from .targets import Targets

__all__ = [
    "ClientName",
    "EventID",
    "Objects",
    "ParameterBuilder",
    "ProgramID",
    "QueryParameter",
    "QueryParams",
    "SkipAndLimit",
    "SubscriptionID",
    "Targets",
    "default_error_handler",
    "get_query",
]



from typing import Literal

from pydantic import Field

from toadr3.models import DocstringBaseModel


class EventPayloadDescriptor(DocstringBaseModel):
    """Contextual information used to interpret event valuesMap values."""

    object_type: Literal["EVENT_PAYLOAD_DESCRIPTOR"] = "EVENT_PAYLOAD_DESCRIPTOR"
    """Used as discriminator."""

    payload_type: str = Field(min_length=1, max_length=128)
    """Enumerated or private string signifying the nature of values."""

    units: str | None = None
    """The measurement units of the payload."""

    currency: str | None = None
    """The currency of the payload, if applicable."""



from typing import Any

from .query_parameter import QueryParameter, QueryParams


class ClientName(QueryParameter):
    """Client name query parameter.

    client_name : str | None
        The client name to filter the items by.
    """

    @staticmethod
    def create_query_parameters(params: QueryParams, args: dict[str, Any]) -> None:
        """Add client_name parameter."""
        if "client_name" in args and args["client_name"] is not None:
            params["clientName"] = args["client_name"]

    @staticmethod
    def check_query_parameters(errors: list[str], args: dict[str, Any]) -> None:
        """Check if client_name parameter is valid."""
        # nothing to check



from typing import Any

from pydantic import AliasGenerator, BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class DocstringBaseModel(BaseModel):
    """BaseModel that supports docstrings for field descriptions."""

    model_config = ConfigDict(
        use_attribute_docstrings=True,
        alias_generator=AliasGenerator(
            validation_alias=to_camel,
            serialization_alias=to_camel,
        ),
        serialize_by_alias=True,
        validate_by_alias=True,
        validate_by_name=True,
    )

    def model_post_init(self, _context: Any, /) -> None:  # noqa: ANN401
        """Call after model class has been initialized."""
        if hasattr(self, "object_type"):
            # We want the discriminator 'objectType' to be part of the JSON during dumping.
            self.model_fields_set.add("object_type")

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"{self.__class__.__name__}({super().__str__()})"



from .docstringbasemodel import DocstringBaseModel
from .event import Event
from .eventpayloaddescriptor import EventPayloadDescriptor
from .interval import Interval
from .intervalperiod import IntervalPeriod
from .objectoperation import ObjectOperation, ObjectType, OperationType
from .problem import Problem
from .program import Program
from .report import Report
from .reportdata import ReportData
from .reportdescriptor import ReportDescriptor
from .reportpayloaddescriptor import ReportPayloadDescriptor
from .subscription import Subscription
from .targettype import TargetType
from .valuesmap import Point, ValuesMap

__all__ = [
    "DocstringBaseModel",
    "Event",
    "EventPayloadDescriptor",
    "Interval",
    "IntervalPeriod",
    "ObjectOperation",
    "ObjectType",
    "OperationType",
    "Point",
    "Problem",
    "Program",
    "Report",
    "ReportData",
    "ReportDescriptor",
    "ReportPayloadDescriptor",
    "Subscription",
    "TargetType",
    "ValuesMap",
]



from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .valuesmap import ValuesMap


class ReportDescriptor(DocstringBaseModel):
    """An object that may be used to request a report from a VEN."""

    payload_type: str = Field(min_length=1, max_length=128)
    """The type of the payload."""

    reading_type: str | None = None
    """Enumerated or private string signifying the type of reading."""

    units: str | None = None
    """The measurement units of the reading."""

    targets: list[ValuesMap] | None = None
    """A list of ValuesMap objects."""

    aggregate: bool = False
    """Whether the report should be aggregated.

    True if report should aggregate results from all targeted resources.
    False if report includes results for each resource.
    """

    start_interval: int = -1
    """The interval at which to generate the report (-1 means end of last)."""

    num_intervals: int = -1
    """The number of intervals to include in a report.

    -1 indicates that all intervals are to be included.
    """

    historical: bool = True
    """Whether the report is historical or not.

    True indicates report on intervals preceding startInterval.
    False indicates report on intervals following startInterval (e.g. forecast).
    """

    frequency: int = -1
    """The frequency of the report.

     Number of intervals that elapse between reports.
    -1 indicates same as numIntervals.
    """

    repeat: int = 1
    """The number of times to repeat the report.

    1 indicates generate one report.
    -1 indicates repeat indefinitely.
    """



import datetime
from typing import Literal

from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .eventpayloaddescriptor import EventPayloadDescriptor
from .interval import Interval
from .intervalperiod import IntervalPeriod
from .reportdescriptor import ReportDescriptor
from .valuesmap import ValuesMap


class Event(DocstringBaseModel):
    """Event object to communicate a Demand Response request to VEN.

    If intervalPeriod is present, sets start time and duration of intervals.
    """

    id: str | None = Field(min_length=1, max_length=128, pattern="^[a-zA-Z0-9_-]*$", default=None)
    """VTN provisioned on object creation."""

    created_date_time: datetime.datetime | None = None
    """VTN provisioned on object creation."""

    modification_date_time: datetime.datetime | None = None
    """VTN provisioned on object modification."""

    object_type: Literal["EVENT"] = "EVENT"
    """VTN provisioned on object creation."""

    program_id: str = Field(
        min_length=1, max_length=128, alias="programID", pattern="^[a-zA-Z0-9_-]*$"
    )
    """ID attribute of program object this event is associated with."""

    event_name: str | None = None
    """User defined string for use in debugging or User Interface."""

    priority: int | None = None
    """Relative priority of event.

    A lower number is a higher priority.
    """
    targets: list[ValuesMap] | None = None
    """A list of valuesMap objects."""

    report_descriptors: list[ReportDescriptor] | None = None
    """A list of reportDescriptor objects.

    Used to request reports from VEN.
    """

    payload_descriptors: list[EventPayloadDescriptor] | None = None
    """A list of payloadDescriptor objects."""

    interval_period: IntervalPeriod | None = None
    """Defines default start and durations of intervals."""

    intervals: list[Interval]
    """List of interval objects."""

    @property
    def created(self) -> datetime.datetime | None:
        """The time the event was created."""
        return self.created_date_time

    @created.setter
    def created(self, value: datetime.datetime | None) -> None:
        """Set the creation date of the event."""
        self.created_date_time = value

    @property
    def modified(self) -> datetime.datetime | None:
        """The time the event was last modified."""
        return self.modification_date_time

    @modified.setter
    def modified(self, value: datetime.datetime | None) -> None:
        """Set the modification date of the event."""
        self.modification_date_time = value



import aiohttp

from ._internal import ParameterBuilder, ProgramID, SkipAndLimit, Targets, get_query
from .access_token import AccessToken
from .models import Event, TargetType

_GET_PARAMS_BUILDER = ParameterBuilder(ProgramID, Targets, SkipAndLimit)


async def get_events(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    program_id: str | None = None,
    target_type: TargetType | str | None = None,
    target_values: list[str] | None = None,
    skip: int | None = None,
    limit: int | None = None,
    extra_params: dict[str, str | int | list[str]] | None = None,
    custom_headers: dict[str, str] | None = None,
) -> list[Event]:
    """Get a list of events from the VTN.

    Lists all the events available from the VTN.
    The results can be filtered by providing a program ID, and/or target type and target values.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url : str
        The URL of the VTN.
    access_token : AccessToken | None
        The access token to use for the request, use None if no token is required.
    program_id : str | None
        The program ID to filter the events by.
    target_type : TargetType | str | None
        The target type to filter the events by.
    target_values : list[str] | None
        The target values to filter the events by (names of the target type).
    skip : int | None
        The number of events to skip (for pagination).
    limit : int | None
        The maximum number of events to return.
    extra_params : dict[str, str | int | list[str]] | None
        Extra query parameters to include in the request.
    custom_headers : dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    list[Event]
        A list of events.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrError
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.
    """
    args = {
        "program_id": program_id,
        "target_type": target_type,
        "target_values": target_values,
        "skip": skip,
        "limit": limit,
    }
    _GET_PARAMS_BUILDER.check_query_parameters(args)
    params = _GET_PARAMS_BUILDER.build_query_parameters(args, extra_params)

    data = await get_query(session, f"{vtn_url}/events", access_token, params, custom_headers)

    if not isinstance(data, list):
        raise ValueError(f"Expected result to be a list. Got {type(data)} instead.")

    result = []
    for event in data:
        result.append(Event.model_validate(event))
    return result



import aiohttp

from toadr3 import AccessToken
from toadr3.models import ObjectType, Subscription, TargetType

from ._internal import (
    ClientName,
    Objects,
    ParameterBuilder,
    ProgramID,
    SkipAndLimit,
    Targets,
    default_error_handler,
    get_query,
)

_GET_PARAMS_BUILDER = ParameterBuilder(ProgramID, ClientName, Targets, Objects, SkipAndLimit)


async def get_subscriptions(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    program_id: str | None = None,
    client_name: str | None = None,
    target_type: TargetType | str | None = None,
    target_values: list[str] | None = None,
    objects: list[str] | list[ObjectType] | None = None,
    skip: int | None = None,
    limit: int | None = None,
    extra_params: dict[str, str | int | list[str]] | None = None,
    custom_headers: dict[str, str] | None = None,
) -> list[Subscription]:
    """List all subscriptions.

    May filter results by programID and clientID as query params.
    May filter results by objects as query param. See objectTypes schema.
    Use skip and pagination query params to limit response size.

    Parameters
    ----------
    session: aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url: str
        The URL of the VTN.
    access_token: AccessToken | None
        The access token to use for the request, use None if no token is required.
    program_id : str | None
        The program ID to filter the subscriptions by.
    client_name : str | None
        The client name to filter the subscriptions by.
    target_type: TargetType | str | None
        The target type to filter the subscriptions by.
    target_values: list[str] | None
        The target values to filter the subscriptions by (names of the target type).
    objects: list[str] | list[ObjectType] | None
        The object types to filter the subscriptions by.
    skip: int | None
        The number of subscriptions to skip (for pagination).
    limit: int | None
        The maximum number of subscriptions to return.
    extra_params: dict[str, str | int | list[str]] | None
        Extra query parameters to include in the request.
    custom_headers: dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    list[Subscription]
        A list of subscriptions.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrException
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.

    """
    args = {
        "client_name": client_name,
        "program_id": program_id,
        "target_type": target_type,
        "target_values": target_values,
        "objects": objects,
        "skip": skip,
        "limit": limit,
    }
    _GET_PARAMS_BUILDER.check_query_parameters(args)
    params = _GET_PARAMS_BUILDER.build_query_parameters(args, extra_params)

    data = await get_query(
        session, f"{vtn_url}/subscriptions", access_token, params, custom_headers
    )

    if not isinstance(data, list):
        raise ValueError(f"Expected result to be a list. Got {type(data)} instead.")

    result = []
    for subscription in data:
        result.append(Subscription.model_validate(subscription))
    return result


async def post_subscription(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    subscription: Subscription,
    custom_headers: dict[str, str] | None = None,
) -> Subscription:
    """Create a new subscription.

    Parameters
    ----------
    session: aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url: str
        The URL of the VTN.
    access_token: AccessToken | None
        The access token to use for the request, use None if no token is required.
    subscription: Subscription
        The subscription to create.
    custom_headers: dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    Subscription
        The subscription object created by the VTN.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrException
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.

    """
    if subscription is None:
        raise ValueError("subscription is required")

    headers: dict[str, str] = {}
    if custom_headers is not None:
        headers |= custom_headers

    if access_token is not None:
        headers["Authorization"] = f"Bearer {access_token.token}"

    vtn_url = vtn_url.rstrip("/")

    data = subscription.model_dump_json(exclude_none=True, exclude_unset=True)
    headers["Content-Type"] = "application/json"

    async with session.post(f"{vtn_url}/subscriptions", headers=headers, data=data) as response:
        if not response.ok:
            match response.status:
                case 400 | 403 | 409 | 500:
                    await default_error_handler(response)
                case _:
                    await default_error_handler(response, "Unexpected error status!")

        data = await response.json()
        return Subscription.model_validate(data)



from typing import Any

from .models import Problem


class ToadrError(Exception):
    """Exception raised for errors in the Toadr API."""

    def __init__(
        self,
        message: str,
        status_code: int | str,
        reason: str | None = None,
        headers: dict[str, Any] | None = None,
        json_response: dict[str, Any] | bytes | None = None,
    ):
        super().__init__(message)
        self._message = message

        if isinstance(status_code, str):
            status_code = int(status_code)

        self._status_code: int = status_code
        if reason is None:
            reason = ""
        self._reason = reason
        if headers is None:
            headers = {}
        self._headers = headers
        if json_response is None:
            json_response = {}
        self._json_response = json_response

        if json_response is not None and isinstance(json_response, dict):
            if "error_description" in json_response:
                self.add_note(json_response["error_description"])
        else:
            self.add_note(str(json_response))

    @property
    def message(self) -> str:
        """A short description of the exception."""
        return self._message

    @property
    def status_code(self) -> int:
        """The HTTP status code of the response."""
        return self._status_code

    @property
    def reason(self) -> str:
        """The reason of the exception (message from aiohttp.ClientResponseError)."""
        return self._reason

    @property
    def headers(self) -> dict[str, Any]:
        """The headers of the response."""
        return self._headers

    @property
    def json_response(self) -> dict[str, Any] | bytes | None:
        """The JSON response from the request (full JSON error response, if available)."""
        return self._json_response

    @classmethod
    def from_problem(cls, problem: Problem, status_code: int | None = None) -> "ToadrError":
        """Create a ToadrError from a `Problem` instance.

        Providing `status_code` will override the status code present in the Problem instance.

        Parameters
        ----------
        problem: Problem
            The problem instance
        status_code: int | None
            Override the HTTP status code present in the Problem instance.
        """
        message = problem.title
        if problem.detail is not None:
            message = f"{message}: {problem.detail}"

        if status_code is None:
            try:
                status_code = int(problem.status)
            except ValueError:
                status_code = 0

        return ToadrError(
            message=message,
            status_code=status_code,
            json_response=problem.model_dump(),
        )



import aiohttp

from toadr3 import AccessToken
from toadr3.models import Program, TargetType

from ._internal import ParameterBuilder, SkipAndLimit, Targets, get_query

_GET_PARAMS_BUILDER = ParameterBuilder(Targets, SkipAndLimit)


async def get_programs(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    target_type: TargetType | str | None = None,
    target_values: list[str] | None = None,
    skip: int | None = None,
    limit: int | None = None,
    extra_params: dict[str, str | int | list[str]] | None = None,
    custom_headers: dict[str, str] | None = None,
) -> list[Program]:
    """Get a list of programs from the VTN.

    List all programs known to the server.
    Use skip and pagination query params to limit response size.

    Parameters
    ----------
    session: aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url: str
        The URL of the VTN.
    access_token: AccessToken | None
        The access token to use for the request, use None if no token is required.
    target_type: TargetType | str | None
        The target type to filter the programs by.
    target_values: list[str] | None
        The target values to filter the programs by (names of the target type).
    skip: int | None
        The number of programs to skip (for pagination).
    limit: int | None
        The maximum number of programs to return.
    extra_params: dict[str, str | int | list[str]] | None
        Extra query parameters to include in the request.
    custom_headers: dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    list[Program]
        A list of programs.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrError
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.
    """
    args = {
        "target_type": target_type,
        "target_values": target_values,
        "skip": skip,
        "limit": limit,
    }
    _GET_PARAMS_BUILDER.check_query_parameters(args)
    params = _GET_PARAMS_BUILDER.build_query_parameters(args, extra_params)

    data = await get_query(session, f"{vtn_url}/programs", access_token, params, custom_headers)

    if not isinstance(data, list):
        raise ValueError(f"Expected result to be a list. Got {type(data)} instead.")

    result = []
    for program in data:
        result.append(Program.model_validate(program))
    return result



import datetime
from typing import Literal

from pydantic import AnyUrl, Field

from .docstringbasemodel import DocstringBaseModel
from .eventpayloaddescriptor import EventPayloadDescriptor
from .intervalperiod import IntervalPeriod
from .reportpayloaddescriptor import ReportPayloadDescriptor
from .valuesmap import ValuesMap


class Program(DocstringBaseModel):
    """Provides program specific metadata from VTN to VEN."""

    id: str | None = Field(min_length=1, max_length=128, pattern="^[a-zA-Z0-9_-]*$", default=None)
    """VTN provisioned on object creation."""

    created_date_time: datetime.datetime | None = None
    """VTN provisioned on object creation."""

    modification_date_time: datetime.datetime | None = None
    """VTN provisioned on object modification."""

    object_type: Literal["PROGRAM"] = "PROGRAM"
    """VTN provisioned on object creation."""

    program_name: str = Field(min_length=1, max_length=128)
    """Short name to uniquely identify program."""

    program_long_name: str | None = None
    """Long name of program for human readability."""

    retailer_name: str | None = None
    """Short name of energy retailer providing the program."""

    retailer_long_name: str | None = None
    """Long name of energy retailer for human readability."""

    program_type: str | None = None
    """A program defined categorization."""

    country: str | None = None
    """Alpha-2 code per ISO 3166-1."""

    principal_subdivision: str | None = None
    """Coding per ISO 3166-2. E.g. state in US."""

    time_zone_offset: datetime.timedelta = datetime.timedelta(seconds=0)
    """Number of hours different from UTC for the standard time applicable to the program."""

    interval_period: IntervalPeriod | None = None
    """The temporal span of the program, could be years long."""

    program_descriptions: list[AnyUrl] | None = None
    """A list of program descriptions."""

    binding_events: bool | None = None
    """True if events are fixed once transmitted."""

    local_price: bool | None = None
    """True if events have been adapted from a grid event."""

    payload_descriptors: list[EventPayloadDescriptor | ReportPayloadDescriptor] | None = None
    """A list of payloadDescriptors."""

    targets: list[ValuesMap] | None = None
    """A list of valuesMap objects."""



from typing import Any

from toadr3.models import TargetType

from .query_parameter import QueryParameter, QueryParams


class Targets(QueryParameter):
    """Target type and target value parameters.

    target_type : TargetType | str | None
        The target type to filter the results by.
    target_values : list[str] | None
        The target values to filter the results by (names of the target type).
    """

    @staticmethod
    def create_query_parameters(params: QueryParams, args: dict[str, Any]) -> None:
        """Add target type and target value parameters."""
        target_type = args.get("target_type")
        target_values = args.get("target_values")

        if target_type is not None and target_values is not None:
            if isinstance(target_type, TargetType):
                params["targetType"] = target_type.value
            else:
                params["targetType"] = target_type  # should be string

            params["targetValues"] = target_values

    @staticmethod
    def check_query_parameters(errors: list[str], args: dict[str, Any]) -> None:
        """Check if skip and limit parameters are valid."""
        target_type = args.get("target_type")
        target_values = args.get("target_values")

        if target_type is not None and target_values is None:
            errors.append("target_values are required when target_type is provided")

        if target_values is not None and target_type is None:
            errors.append("target_type is required when target_values are provided")

        if target_values is not None and not isinstance(target_values, list):
            errors.append("target_values must be a list of strings")

        if target_type is not None and not isinstance(target_type, (TargetType, str)):
            errors.append("target_type must be TargetType or str")



from typing import Literal

from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .reportdescriptor import ReportDescriptor


class ReportPayloadDescriptor(DocstringBaseModel):
    """Contextual information used to interpret report payload values."""

    object_type: Literal["REPORT_PAYLOAD_DESCRIPTOR"] = "REPORT_PAYLOAD_DESCRIPTOR"
    """Used as discriminator."""

    payload_type: str = Field(min_length=1, max_length=128)
    """Enumerated or private string signifying the nature of values."""

    reading_type: str | None = None
    """Enumerated or private string signifying the type of reading."""

    units: str | None = None
    """Units of measure."""

    accuracy: float | None = None
    """A quantification of the accuracy of a set of payload values."""

    confidence: int | None = Field(ge=0, le=100, default=None)
    """A quantification of the confidence of a set of payload values."""

    @staticmethod
    def from_report_descriptor(report_descriptor: ReportDescriptor) -> "ReportPayloadDescriptor":
        """Create a ReportPayloadDescriptor from a ReportDescriptor."""
        return ReportPayloadDescriptor(
            payload_type=report_descriptor.payload_type,
            reading_type=report_descriptor.reading_type,
            units=report_descriptor.units,
        )



from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .interval import Interval
from .intervalperiod import IntervalPeriod


class ReportData(DocstringBaseModel):
    """Report data associated with a resource."""

    resource_name: str = Field(min_length=1, max_length=128)
    """The name of the resource.

    A value of AGGREGATED_REPORT indicates an aggregation of more that one resource's data.
    """

    interval_period: IntervalPeriod | None = None
    """Defines default start and durations of intervals."""

    intervals: list[Interval]
    """List of interval objects."""



from abc import ABC, abstractmethod
from typing import Any, TypeAlias

QueryParams: TypeAlias = dict[str, str | int | list[str]]


class QueryParameter(ABC):
    """QueryParameter base class."""

    @staticmethod
    @abstractmethod
    def create_query_parameters(params: QueryParams, args: dict[str, Any]) -> None:
        """Process incoming arguments into query parameters."""

    @staticmethod
    @abstractmethod
    def check_query_parameters(errors: list[str], args: dict[str, Any]) -> None:
        """Check if query parameter is valid."""



from pydantic import HttpUrl

from toadr3.models import DocstringBaseModel


class Problem(DocstringBaseModel):
    """Reusable error response."""

    type: HttpUrl | None = None
    """An absolute URI that identifies the problem type.

    When dereferenced, it SHOULD provide human-readable documentation for the problem type
    (e.g., using HTML).
    """

    title: str
    """A short, summary of the problem type.

    Written in english and readable for engineers (usually not suited for non
    technical stakeholders and not localized); example: Service Unavailable.
    """

    status: int
    """The HTTP status code.

    Generated by the origin server for this occurrence of the problem.
    """

    detail: str | None = None
    """A human readable explanation specific to this occurrence of the problem."""

    instance: HttpUrl | None = None
    """An absolute URI that identifies the specific occurrence of the problem.

    It may or may not yield further information if dereferenced.
    """



from . import models
from .access_token import (
    AccessToken,
    OAuthAudienceConfig,
    OAuthConfig,
    OAuthScopeConfig,
    acquire_access_token,
    acquire_access_token_from_config,
)
from .client import ToadrClient
from .events import get_events
from .exceptions import ToadrError
from .programs import get_programs
from .reports import get_reports, post_report
from .subscriptions import get_subscriptions, post_subscription

__all__ = [
    "AccessToken",
    "OAuthAudienceConfig",
    "OAuthConfig",
    "OAuthScopeConfig",
    "ToadrClient",
    "ToadrError",
    "acquire_access_token",
    "acquire_access_token_from_config",
    "get_events",
    "get_programs",
    "get_reports",
    "get_subscriptions",
    "models",
    "post_report",
    "post_subscription",
]



import datetime
import os

import aiohttp
from aiohttp import ClientResponse

from .exceptions import ToadrError


class OAuthConfig:
    """Model encapsulating values required for authorization."""

    def __init__(
        self,
        token_url: str,
        grant_type: str,
        claims: dict[str, str],
        client_id: str | None = None,
        client_secret: str | None = None,
    ) -> None:
        """Encapsulate configuration required for authorization.

        `client_id` and `client_secret` can be None if they are available
        in the environment as CLIENT_ID and CLIENT_SECRET.

        Parameters
        ----------
        session : aiohttp.ClientSession
            The aiohttp session to use for the request.
        token_url : str
            The URL to acquire the token from.
        grant_type : str
            The grant type to use for the token request.
        claims : dict[str, str]
            Collection of claims to include in the token request.
        client_id : str | None
            The client ID or None if acquirable from the environment as CLIENT_ID.
        client_secret : str | None
            The client secret or None if acquirable from the environment as CLIENT_SECRET.
        """
        self._token_url = token_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._grant_type = grant_type
        self._claims = claims

    @property
    def url(self) -> str:
        """URL to the OAuth server."""
        return self._token_url

    @property
    def client_id(self) -> str | None:
        """Client ID."""
        return self._client_id

    @property
    def client_secret(self) -> str | None:
        """Client secret."""
        return self._client_secret

    @property
    def grant_type(self) -> str:
        """Grant type."""
        return self._grant_type

    @property
    def claims(self) -> dict[str, str]:
        """Claims to include in the token request."""
        return self._claims


class OAuthScopeConfig(OAuthConfig):
    """Model encapsulating values required for authorization with scope."""

    def __init__(
        self,
        token_url: str,
        grant_type: str,
        scope: str,
        client_id: str | None = None,
        client_secret: str | None = None,
    ) -> None:
        """Encapsulate configuration required for authorization with scope.

        `client_id` and `client_secret` can be None if they are available
        in the environment as CLIENT_ID and CLIENT_SECRET.

        Parameters
        ----------
        token_url : str
            The URL to acquire the token from.
        grant_type : str
            The grant type to use for the token request.
        scope : str
            The scope of the token.
        client_id : str | None
            The client ID or None if acquirable from the environment as CLIENT_ID.
        client_secret : str | None
            The client secret or None if acquirable from the environment as CLIENT_SECRET.
        """
        if scope is None:
            raise ValueError("scope cannot be None")
        super().__init__(
            token_url,
            grant_type,
            claims={"scope": scope},
            client_id=client_id,
            client_secret=client_secret,
        )

    @property
    def scope(self) -> str:
        """Scope."""
        return self._claims["scope"]


class OAuthAudienceConfig(OAuthConfig):
    """Model encapsulating values required for authorization with audience."""

    def __init__(
        self,
        token_url: str,
        grant_type: str,
        audience: str,
        client_id: str | None = None,
        client_secret: str | None = None,
    ) -> None:
        """Encapsulate configuration required for authorization with audience.

        `client_id` and `client_secret` can be None if they are available
        in the environment as CLIENT_ID and CLIENT_SECRET.

        Parameters
        ----------
        token_url : str
            The URL to acquire the token from.
        grant_type : str
            The grant type to use for the token request.
        audience : str
            The audience of the token.
        client_id : str | None
            The client ID or None if acquirable from the environment as CLIENT_ID.
        client_secret : str | None
            The client secret or None if acquirable from the environment as CLIENT_SECRET.
        """
        if audience is None:
            raise ValueError("audience cannot be None")
        super().__init__(
            token_url,
            grant_type,
            claims={"audience": audience},
            client_id=client_id,
            client_secret=client_secret,
        )

    @property
    def audience(self) -> str:
        """Audience."""
        return self._claims["audience"]


class AccessToken:
    """Access token object.

    Enables tracking of expiration time and checking if the token is expired.
    """

    def __init__(self, token: str, expires_in: int):
        """Initialize the access token with an expiration time in seconds.

        Checking if the token is expired can be done with the is_expired() method. The
        is_expired() method will return True if the token is expired or will expire within
        the next 60 seconds.

        Parameters
        ----------
        token : str
            The access token.
        expires_in : int
            The time in seconds until the token expires.
        """
        self._token = token
        time_delta = datetime.timedelta(seconds=expires_in)
        self._expires_at = datetime.datetime.now(tz=datetime.timezone.utc) + time_delta

    @property
    def token(self) -> str:
        """The access token."""
        return self._token

    @property
    def expires_at(self) -> datetime.datetime:
        """The time in seconds until the token expires."""
        return self._expires_at

    @property
    def expires_in(self) -> int:
        """The time in seconds until the token expires."""
        expires_in = self._expires_at - datetime.datetime.now(tz=datetime.timezone.utc)
        return int(expires_in.total_seconds())

    def is_expired(self) -> bool:
        """Check if the token is expired."""
        time_delta = self._expires_at - datetime.datetime.now(tz=datetime.timezone.utc)
        return time_delta < datetime.timedelta(seconds=60)

    def __str__(self) -> str:
        """Return a string representation of the access token."""
        return f"{self.token}"

    def __repr__(self) -> str:
        """Return a string representation of the access token."""
        return f"AccessToken(token='{self.token}', expires_in={self.expires_in})"


async def acquire_access_token_from_config(
    session: aiohttp.ClientSession, config: OAuthConfig
) -> AccessToken:
    """Acquire an access token from the token provider.

    Connect to the token provider and acquire an access token. The access token will be returned
    as an AccessToken object. `client_id` and `client_secret` can be None if they are available
    in the environment as CLIENT_ID and CLIENT_SECRET.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    config: OAuthConfig
        The configuration object required to acquire an access token. It can be either an instance
        of OAuthScopeConfig or OAuthAudienceConfig.

    Returns
    -------
    AccessToken
        The access token object.

    Raises
    ------
    ValueError
        If the `client_id` or `client_secret` are not provided and not available in the environment.
        If 'claims' are not provided or empty.
    toadr3.ToadrError
        If the request to the token provider fails.
    """
    return await acquire_access_token(
        session,
        config.url,
        config.grant_type,
        config.claims,
        config.client_id,
        config.client_secret,
    )


async def acquire_access_token(
    session: aiohttp.ClientSession,
    token_url: str,
    grant_type: str,
    claims: dict[str, str],
    client_id: str | None = None,
    client_secret: str | None = None,
) -> AccessToken:
    """Acquire an access token from the token provider.

    Connect to the token provider and acquire an access token. The access token will be returned
    as an AccessToken object. `client_id` and `client_secret` can be None if they are available
    in the environment as CLIENT_ID and CLIENT_SECRET.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    token_url : str
        The URL to acquire the token from.
    grant_type : str
        The grant type to use for the token request.
    claims: dict[str, str]
        Collection of claims to include in the token request.
    client_id : str | None
        The client ID or None if acquirable from environment as CLIENT_ID.
    client_secret : str | None
        The client secret or None if acquirable from environment as CLIENT_SECRET.

    Returns
    -------
    AccessToken
        The access token object.

    Raises
    ------
    ValueError
        If the `client_id` or `client_secret` are not provided and not available in the environment.
    toadr3.ToadrError
        If the request to the token provider fails.
    """
    if client_id is None:
        if "CLIENT_ID" not in os.environ:
            raise ValueError("client_id is required")
        client_id = os.getenv("CLIENT_ID")

    if client_secret is None:
        if "CLIENT_SECRET" not in os.environ:
            raise ValueError("client_secret is required")
        client_secret = os.getenv("CLIENT_SECRET")

    if grant_type is None:
        raise ValueError("grant_type is required")

    if claims is None:
        raise ValueError("claims are required")

    if len(claims) == 0:
        raise ValueError("claims cannot be empty")

    credentials = {
        "grant_type": grant_type,
        "client_id": client_id,
        "client_secret": client_secret,
    }

    credentials |= claims

    async with session.post(token_url, data=credentials) as response:
        # 400 is the start of HTTP error codes
        if response.status >= 400:  # noqa PLR2004 - Magic value used in comparison
            await _process_error(response)  # will raise a ToadrError

        data = await response.json()
        return AccessToken(data["access_token"], data["expires_in"])


async def _process_error(response: ClientResponse) -> None:
    message = "Failed to acquire access token: "
    json = await response.json()

    if "error" in json:
        message += json["error"]
    else:
        message += json

    raise ToadrError(
        message,
        status_code=response.status,
        reason=response.reason,
        headers=response.headers,  # type: ignore[arg-type]
        json_response=json,
    )



import datetime
from typing import Any, Literal

from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .event import Event
from .interval import Interval
from .reportdata import ReportData
from .reportpayloaddescriptor import ReportPayloadDescriptor
from .targettype import TargetType
from .valuesmap import ValuesMap


class Report(DocstringBaseModel):
    """Report object for a report."""

    id: str | None = Field(min_length=1, max_length=128, pattern="^[a-zA-Z0-9_-]*$", default=None)
    """VTN provisioned on object creation."""

    created_date_time: datetime.datetime | None = None
    """VTN provisioned on object creation."""

    modification_date_time: datetime.datetime | None = None
    """VTN provisioned on object modification."""

    object_type: Literal["REPORT"] = "REPORT"
    """VTN provisioned on object creation."""

    program_id: str = Field(
        min_length=1, max_length=128, alias="programID", pattern="^[a-zA-Z0-9_-]*$"
    )
    """ID attribute of program object this report is associated with."""

    event_id: str = Field(min_length=1, max_length=128, alias="eventID", pattern="^[a-zA-Z0-9_-]*$")
    """ID attribute of event object this report is associated with."""

    client_name: str = Field(
        min_length=1, max_length=128, alias="clientName", pattern="^[a-zA-Z0-9_-]*$"
    )
    """User generated identifier; may be VEN ID provisioned during program enrollment."""

    report_name: str | None = None
    """User defined string for use in debugging or User Interface."""

    payload_descriptors: list[ReportPayloadDescriptor] | None = None
    """An optional list of objects that provide context to payload types."""

    resources: list[ReportData]

    @property
    def created(self) -> datetime.datetime | None:
        """The time the event was created."""
        return self.created_date_time

    @created.setter
    def created(self, value: datetime.datetime | None) -> None:
        """Set the creation date of the report."""
        self.created_date_time = value

    @property
    def modified(self) -> datetime.datetime | None:
        """The time the event was last modified."""
        return self.modification_date_time

    @modified.setter
    def modified(self, value: datetime.datetime | None) -> None:
        """Set the modification date of the report."""
        self.modification_date_time = value

    @staticmethod
    def create_report(
        event: Event,
        client_name: str,
        report_type: str,
        report_values: list[Any],
        report_name: str | None = None,
        target_type: TargetType | str = TargetType.RESOURCE_NAME,
    ) -> "Report":
        """Create a new report object.

        This will create a single report object for the given event for the specific report types.
        There are several assumptions made in this function:
        1. You want/can only create a report for one report descriptor (at a time).
        2. There is only one target the report applies to.
        3. There is only one interval, and it is specified in an 'intervalPeriod' at the root level.

        Parameters
        ----------
        event : Event
            The event object to create the report for.
        client_name : str
            The client name for the report.
        report_type : str
            The report type for the report (payload type of one of report descriptors).
        report_values : list[Any]
            The report values for the report.
        report_name : str, optional
            The name for the report (for debugging).
        target_type : TargetType | str, optional
            The type of target to look for, the default being TargetType.RESOURCE_NAME.

        Raises
        ------
        ValueError
            For missing and invalid arguments.

        Returns
        -------
        Report
            The created report object.
        """
        if not report_type or not isinstance(report_type, str):
            raise ValueError("report_type is required.")

        if not report_values or not isinstance(report_values, list):
            raise ValueError("report_values is required.")

        if event.targets is None or len(event.targets) == 0:
            raise ValueError("event does not have any targets.")

        if isinstance(target_type, TargetType):
            target_type = target_type.value

        found_resource_name = False
        for target in event.targets:
            if target.type == target_type:
                found_resource_name = True
                break

        if not found_resource_name:
            raise ValueError(f"event does not have a target for type {target_type}.")

        if event.report_descriptors is None or len(event.report_descriptors) == 0:
            raise ValueError("event does not have any report_descriptors.")

        report_descriptor = None
        for rd in event.report_descriptors:
            if rd.payload_type == report_type:
                report_descriptor = rd
                break

        if report_descriptor is None:
            raise ValueError(f"event does not have a report_descriptor for {report_type}.")

        return Report(
            program_id=event.program_id,
            event_id=event.id,
            client_name=client_name,
            report_name=report_name,
            payload_descriptors=[
                ReportPayloadDescriptor.from_report_descriptor(event.report_descriptors[0])
            ],
            resources=[
                ReportData(
                    resource_name=str(event.targets[0].values[0]),
                    interval_period=event.interval_period,
                    intervals=[
                        Interval(
                            id=0,
                            payloads=[
                                ValuesMap(
                                    type=report_type,
                                    values=report_values,
                                )
                            ],
                        )
                    ],
                )
            ],
        )



import datetime
import re


def parse_iso8601_duration(duration: str) -> datetime.timedelta:
    """Parse an ISO 8601 duration string into a timedelta object.

    Since Python's timedelta does not support years, months and weeks, this function will fail if
    those parts of the duration are specified. In addition, only seconds can have a fraction part.

    We have added support for weeks by adding 7 days to the days part of the timedelta.

    Months and years are not supported because they have variable lengths and are not suitable for
    conversion to a fixed number of days without a reference date.
    """
    dec_or_int = r"((\d+[.]\d+)|(\d+))"  # decimal or integer
    pattern = (
        r"^(?:(?P<negative>[-]?))?"  # negative sign is optionally at the start of the string
        r"P(?!$)"  # P is required to be at the start of the string
        r"(?:(?P<weeks>[-]?\d+)W)?"  # weeks are optional
        r"(?:(?P<days>[-]?\d+)D)?"  # days are optional
        rf"(?:T(?=[-]?{dec_or_int}[HMS])"  # T is required if time part is present (+ look ahead)
        r"(?:(?P<hours>[-]?\d+)H)?"  # hours are optional
        r"(?:(?P<minutes>[-]?\d+)M)?"  # minutes are optional
        rf"(?:(?P<seconds>[-]?{dec_or_int})S)?"  # seconds are optional (can be fractional)
        r")?$"  # end of string
    )
    match = re.match(pattern, duration)
    if not match:
        raise ValueError(f"Invalid ISO 8601 duration: {duration}")

    factor = 1
    parts = {}
    for k, v in match.groupdict("0").items():
        if k == "negative":
            if v == "-":
                factor = -1
        else:
            parts[k] = float(v) * factor

    if "weeks" in parts:
        if "days" not in parts:
            parts["days"] = 0
        parts["days"] += parts.pop("weeks") * 7

    return datetime.timedelta(**parts)


def create_iso8601_duration(delta: datetime.timedelta) -> str:
    """Create an ISO 8601 duration string from a timedelta object.

    If the timedelta is negative, the duration will be prefixed with a negative sign.

    Parameters
    ----------
    delta : datetime.timedelta
        The timedelta object to convert to an ISO 8601 duration string.

    Returns
    -------
    str
        The ISO 8601 duration string.
    """
    seconds = delta.total_seconds()

    if seconds == 0:
        return "PT0S"

    negative = ""
    if seconds < 0:
        negative = "-"
        seconds = abs(seconds)

    weeks, seconds = divmod(seconds, 7 * 24 * 3600)
    days, seconds = divmod(seconds, 24 * 3600)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)

    duration = f"{negative}P"
    if weeks != 0:
        duration += f"{int(weeks)}W"

    if days != 0:
        duration += f"{int(days)}D"

    if hours != 0 or minutes != 0 or seconds != 0:
        duration += "T"

    if hours != 0:
        duration += f"{int(hours)}H"

    if minutes != 0:
        duration += f"{int(minutes)}M"

    if seconds != 0:
        if seconds.is_integer():
            duration += f"{int(seconds)}S"
        else:
            num = f"{seconds:f}".rstrip("0").rstrip(".")
            duration += f"{num}S"
    return duration



from pydantic import Field

from toadr3.models import DocstringBaseModel


class Point(DocstringBaseModel):
    """A pair of floats typically used as a point on a 2 dimensional grid."""

    x: float
    """The x coordinate of the point."""

    y: float
    """The y coordinate of the point."""


class ValuesMap(DocstringBaseModel):
    """Represents one or more values associated with a type."""

    type: str = Field(min_length=1, max_length=128)
    """Enumerated or private string signifying the nature of values."""

    values: list[int | float | str | bool | Point]
    """A list of data points. Most often a singular value such as a price."""



import re
from typing import Any

from .query_parameter import QueryParameter, QueryParams

_regex = re.compile(r"^[a-zA-Z0-9_-]*$")


class ObjectID(QueryParameter):
    """Object ID query parameter.

    Represents an object ID, for example program_id or event_id.
    """

    _attribute = ("unset", "unset")
    """Override this to provide class specific values.

    The first version represent the input argument, for example: 'value_id'
    The second value represent the query parameter, for example: 'valueID'
    """

    _nullable = True
    """Override this to disable nullable values."""

    @classmethod
    def create_query_parameters(cls, params: QueryParams, args: dict[str, Any]) -> None:
        """Add a parameter that is an object ID."""
        arg = cls._attribute[0]
        if arg in args and args[arg] is not None:
            params[cls._attribute[1]] = args[arg]

    @classmethod
    def check_query_parameters(cls, errors: list[str], args: dict[str, Any]) -> None:
        """Check if the ObjectID parameter is valid."""
        arg = cls._attribute[0]
        object_id = args[arg]

        if not cls._nullable and object_id is None:
            errors.append(f"{arg} cannot be None")
            return

        if object_id is not None:
            if not isinstance(object_id, str):
                errors.append(f"{arg} must be a string")
                return

            if not 1 <= len(object_id) <= 128:  # noqa: PLR2004
                errors.append(f"{arg} must be between 1 and 128 characters long")

            if _regex.match(object_id) is None:
                msg = f"{arg} '{object_id}' does not match regex '{_regex.pattern}'"
                errors.append(msg)


class ProgramID(ObjectID):
    """Program ID query parameter.

    program_id : str | None
        The program ID to filter the items by.
    """

    _attribute = ("program_id", "programID")


class EventID(ObjectID):
    """Event ID query parameter.

    event_id : str | None
        The event ID to filter the items by.
    """

    _attribute = ("event_id", "eventID")


class SubscriptionID(ObjectID):
    """Subscription ID query parameter."""

    _attribute = ("subscription_id", "subscriptionID")
    _nullable = False



import datetime
from typing import Literal

from pydantic import Field

from .docstringbasemodel import DocstringBaseModel
from .objectoperation import ObjectOperation
from .valuesmap import ValuesMap


class Subscription(DocstringBaseModel):
    """An object created by a client to receive notification of operations on objects.

    Clients may subscribe to be notified when a type of object is created,
    updated, or deleted.
    """

    id: str | None = Field(min_length=1, max_length=128, pattern="^[a-zA-Z0-9_-]*$", default=None)
    """VTN provisioned on object creation."""

    created_date_time: datetime.datetime | None = None
    """VTN provisioned on object creation."""

    modification_date_time: datetime.datetime | None = None
    """VTN provisioned on object modification."""

    object_type: Literal["SUBSCRIPTION"] = "SUBSCRIPTION"
    """Used as discriminator."""

    client_name: str = Field(min_length=1, max_length=128)
    """User generated identifier; may be VEN ID provisioned during program enrollment."""

    program_id: str = Field(
        min_length=1, max_length=128, alias="programID", pattern="^[a-zA-Z0-9_-]*$"
    )
    """ID attribute of program object this event is associated with."""

    object_operations: list[ObjectOperation]
    """List of objects and operations to subscribe to."""

    targets: list[ValuesMap] | None = None
    """A list of valuesMap objects.

    Used by server to filter callbacks.
    """

    @property
    def created(self) -> datetime.datetime | None:
        """The time the event was created."""
        return self.created_date_time

    @created.setter
    def created(self, value: datetime.datetime | None) -> None:
        """Set the creation date of the event."""
        self.created_date_time = value

    @property
    def modified(self) -> datetime.datetime | None:
        """The time the event was last modified."""
        return self.modification_date_time

    @modified.setter
    def modified(self, value: datetime.datetime | None) -> None:
        """Set the modification date of the event."""
        self.modification_date_time = value



from typing import Any

import aiohttp

from toadr3.access_token import AccessToken
from toadr3.exceptions import ToadrError


async def default_error_handler(response: aiohttp.ClientResponse, msg: str | None = None) -> None:
    """Error handler that generates a ToadrError from a Problem JSON response.

    Parameters
    ----------
    response: aiohttp.ClientResponse
        Response from server
    msg: str | None
        Pre-message to add to the message.

    Raises
    ------
    ToadrError
        Raises a ToadrError from a Problem JSON response.
    """
    json_response = await response.json()  # JSON should be of type Problem schema
    status = response.status

    message = ""
    if msg is not None:
        message = f"{msg} "

    message += json_response["title"]

    if "detail" in json_response:
        message = f"{message} {status} - {json_response['detail']}"

    raise ToadrError(
        message,
        status_code=response.status,
        reason=response.reason,
        headers=response.headers,  # type: ignore[arg-type]
        json_response=json_response,
    )


async def get_query(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    params: dict[str, str | int | list[str]] | None = None,
    custom_headers: dict[str, str] | None = None,
) -> Any:  # noqa: ANN401
    """Perform default handling of a GET query.

    Parameters
    ----------
    session: aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url: str
        The URL of the VTN.
    access_token: AccessToken | None
        The access token to use for the request, use None if no token is required.
    params: dict[str, str | int | list[str]] | None
        Extra query parameters to include in the request.
    custom_headers: dict[str, str] | None
        Extra headers to include in the request.

    Returns
    -------
    Any
        The response parsed as JSON.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrException
        If the request to the VTN fails. Specifically, if the response status is 400, 403, or 500,
    aiohttp.ClientError
        If there is an unexpected error with the HTTP request to the VTN.

    """
    headers: dict[str, str] = {}
    if custom_headers is not None:
        headers |= custom_headers

    if access_token is not None:
        headers["Authorization"] = f"Bearer {access_token.token}"

    vtn_url = vtn_url.rstrip("/")

    async with session.get(vtn_url, params=params, headers=headers) as response:
        if not response.ok:
            match response.status:
                case 400 | 403 | 500:
                    await default_error_handler(response)
                case _:
                    await default_error_handler(response, "Unexpected error status!")

        return await response.json()
